"""The ``utils`` submodule provides several useful functions for computes, annotations, detection and tracking.

This submodule contains the following class:

- ``Display`` : Display tracking.
- ``FramesWithInfo`` : Frames with information.
- ``GaussianMixture`` : This class contains several methods to compute gaussian Probability Density Function.

This submodule contains the following functions:

- ``exp_and_normalize(lw:ndarray[Any,dtype[float64]])`` : This fonction transforms a weight vector / matrix into a normalized gaussian vector / matrix.
- ``gather_filenames_for_video_in_annotations(video:Dict, images:List[Dict], data_dir:str)`` : Gather image names from a video for annotations.
- ``generate_video_with_annotations(reader:Any,output_detected:Dict,output_filename:str,skip_frames:int,downscale:float,
    logger:Logger,gps_data:Optional[ndarray]=None,labels2icons:Optional[ndarray]=None)`` : Generates output video at 24 fps, with optional ``gps_data``.
- ``get_detections_for_video(reader:Any, detector:Any, batch_size:int=16, device:Optional[str]=None)`` : Get detections for video.
- ``in_frame(position:Union[Tuple[int,int],List[int,int],array[int,int]],
                shape:Union[Tuple[int,int],List[int,int],array[int,int]],
                border:float=0.02)`` : Check if the (object) position is inside the image.
- ``overlay_transparent(background:ndarray, overlay:ndarray, x:int, y:int)`` : Overlays a transparent image over a background at topleft corner (x,y).
- ``read_tracking_results(input_file:str) `` : Read the input filename and interpret it as tracklets.
- ``resize_external_detections(detections:ndarray, ratio:float)`` : Resize external detections with specific ratio.
- ``write_tracking_results_to_file(results:ndarray, ratio_x:float, ratio_y:float, output_filename:str)`` : Writes the output results of a tracking

"""

import os
from collections import defaultdict
from time import time
from logging import Logger
from typing import Any, Iterable, Union, Tuple, List, Dict, Optional

import cv2
from cv2 import Mat
import matplotlib.pyplot as plt
import numpy as np
from numpy import array, generic, ndarray, dtype, float64

import torch
from scipy.stats import multivariate_normal
from torch.utils.data import DataLoader
from skvideo.io import FFmpegWriter
from skimage.transform import downscale_local_mean

from plasticorigins.detection.transforms import TransformFrames
from plasticorigins.tools.video_readers import TorchIterableFromReader


class GaussianMixture:

    """This class contains several methods to compute gaussian Probability Density Function (pdf),
    the standard LOGistic Probability Density Function (logpdf) and the Cumulative Distribution Function (cdf).

    Args:
        means (ndarray[Any,dtype[float64]]]): the mean vector of the gaussian model
        covariance (ndarray[Any,dtype[float64]]]): the covariance matrix of the gaussian model
        weights (ndarray[Any,dtype[float64]]]): the weights of the predictive model
    """

    def __init__(
        self,
        means: ndarray[Any, dtype[float64]],
        covariance: ndarray[Any, dtype[float64]],
        weights: ndarray[Any, dtype[float64]],
    ):
        self.components = [
            multivariate_normal(mean=mean, cov=covariance) for mean in means
        ]
        self.weights = weights

    def pdf(self, x: ndarray) -> ndarray:

        """This method computes the Probability Density Function (pdf) for Gaussian Mixture.

        Args:
            x (ndarray[Any]): the input vector

        Returns:
            result (ndarray[Any,dtype[float64]]) : the Probability Density Function for the input vector x
        """

        result = 0
        for weight, component in zip(self.weights, self.components):
            result += weight * component.pdf(x)
        return result

    def logpdf(self, x: ndarray) -> ndarray:

        """This method computes the standard LOGistic (i.e, mean=0, std=π/sqrt(3)) Probability Density Function (logpdf) for Gaussian Mixture.

        Args:
            x (ndarray[Any]): the input vector

        Returns:
            result (ndarray[Any,dtype[float64]]) : the standard LOGistic Probability Density Function for the input vector x
        """

        return np.log(self.pdf(x))

    def cdf(self, x: ndarray) -> ndarray:

        """This method computes the Cumulative Density Function (pdf) for Gaussian Mixture.

        Args:
            x (ndarray[Any]): the input vector

        Returns:
            result (ndarray[Any,dtype[float64]]) : the Cumulative Density Function for the input vector x
        """

        result = 0
        for weight, component in zip(self.weights, self.components):
            result += weight * component.cdf(x)
        return result


def exp_and_normalize(lw: ndarray[Any, dtype[float64]]) -> ndarray[Any, dtype[float64]]:

    """This fonction transforms a weight vector / matrix into a normalized gaussian vector / matrix.

    Args:
        lw (ndarray[Any,dtype[float64]]): the input vector / matrix to normalize

    Returns:
        w (ndarray[Any,dtype[float64]]): the normalized gaussian weight vector / matrix
    """

    w = np.exp(lw - lw.max())

    return w / w.sum()


def in_frame(
    position: Union[Tuple[int, int], List[int], array],
    shape: Union[Tuple[int, int], List[int], array],
    border: float = 0.02,
) -> bool:

    """Check if the (object) position is inside the image.

    Args:
        position (Union[Tuple[int,int],List[int,int],array[int,int]]): the ``(x, y)`` position of the object
        shape (Union[Tuple[int,int],List[int,int],array[int,int]]): the frame shape ``(H, W)``
        border (float): the thickness of the image border. Set as default to ``0.02``.

    Returns:
        in_frame (bool): ``True`` if the object is in the frame, ``False`` if not

    """

    shape_x = shape[1]
    shape_y = shape[0]
    x = position[0]
    y = position[1]

    return (
        x > border * shape_x
        and x < (1 - border) * shape_x
        and y > border * shape_y
        and y < (1 - border) * shape_y
    )


def gather_filenames_for_video_in_annotations(
    video: Dict, images: List[Dict], data_dir: str
) -> List[str]:

    """Gather image names from a video for annotations.

    Args:
        video (Dict): the input video with its `id` key
        images (List[Dict]): the list of all images
        data_dir (str): the data path for images

    Returns:
        list_img_names (List[Any,type[str]]): the list of the image paths from data directory
    """

    images_for_video = [image for image in images if image["video_id"] == video["id"]]
    images_for_video = sorted(images_for_video, key=lambda image: image["frame_id"])

    return [os.path.join(data_dir, image["file_name"]) for image in images_for_video]


def get_detections_for_video(
    reader: Any,
    detector: Any,
    batch_size: int = 16,
    device: Optional[str] = None,
) -> List[array]:

    """Get detections for video.

    Args:
        reader (Any): the video reader used
        detector (Any): the detector for video
        batch_size (int): the batch size. Set as default to ``16``.
        device (Optional[str]): the device used ("cpu", "cuda",...). Set as default to ``None``.

    Returns:
        detections (List[array]): the list of built detections from the input video
    """

    detections = []
    dataset = TorchIterableFromReader(reader, TransformFrames())
    loader = DataLoader(dataset, batch_size=batch_size)
    average_times = []
    with torch.no_grad():
        for preprocessed_frames in loader:
            time0 = time()
            detections_for_frames = detector(preprocessed_frames.to(device))
            average_times.append(time() - time0)
            for detections_for_frame in detections_for_frames:
                if len(detections_for_frame):
                    detections.append(detections_for_frame)
                else:
                    detections.append(np.array([]))
    print(f"Frame-wise inference time: {batch_size/np.mean(average_times)} fps")
    return detections


def overlay_transparent(
    background: ndarray, overlay: ndarray, x: int, y: int
) -> ndarray[Any, dtype[float64]]:

    """Overlays a transparent image over a background at topleft corner (x,y).

    Args:
        background (ndarray): the input background
        overlay (ndarray): the input overlay
        x (int): the x-position of the topleft corner
        y (int): the y-position of the topleft corner

    Returns:
        background (ndarray[Any,dtype[float64]]): the transformed background
    """

    background_width = background.shape[1]
    background_height = background.shape[0]

    if x >= background_width or y >= background_height:
        return background

    h, w = overlay.shape[0], overlay.shape[1]

    if x + w > background_width:
        w = background_width - x
        overlay = overlay[:, :w]

    if y + h > background_height:
        h = background_height - y
        overlay = overlay[:h]

    if overlay.shape[2] < 4:
        overlay = np.concatenate(
            [
                overlay,
                np.ones(
                    (overlay.shape[0], overlay.shape[1], 1),
                    dtype=overlay.dtype,
                )
                * 255,
            ],
            axis=2,
        )

    overlay_image = overlay[..., :3]
    mask = overlay[..., 3:] / 255.0

    background[y : y + h, x : x + w] = (1.0 - mask) * background[
        y : y + h, x : x + w
    ] + mask * overlay_image

    return background


def generate_video_with_annotations(
    reader: Any,
    output_detected: dict,
    output_filename: str,
    skip_frames: int,
    downscale: float,
    logger: Logger,
    gps_data: Optional[ndarray] = None,
    labels2icons: Optional[ndarray] = None,
) -> None:

    """Generates output video at 24 fps, with optional ``gps_data``.

    Args:
        reader (Any): the video reader used
        output_detected (Dict): the dictionnary of detected trashs
        output_filename (str): the name of the output file
        skip_frames (int): the frequence to skip frames
        downscale (float): the downscale factor
        logger (Logger): the logger to use
        gps_data (Optional[ndarray]): the gps data. Set as default to ``None``.
        labels2icons (Optional[ndarray]): the multi-dimensionnal array with labels to icons. Set as default to ``None``.
    """

    fps = 24
    logger.info("---Intepreting json")
    results = defaultdict(list)
    for trash in output_detected["detected_trash"]:
        for k, v in trash["frame_to_box"].items():
            frame_nb = int(k) - 1
            object_nb = trash["id"] + 1
            object_class = trash["label"]
            center_x = v[0]
            center_y = v[1]
            results[frame_nb * (skip_frames + 1)].append(
                (object_nb, center_x, center_y, object_class)
            )
            # append next skip_frames
            if str(frame_nb + 2) in trash["frame_to_box"]:
                next_trash = trash["frame_to_box"][str(frame_nb + 2)]
                next_x = next_trash[0]
                next_y = next_trash[1]
                for i in range(1, skip_frames + 1):
                    new_x = center_x + (next_x - center_x) * i / (skip_frames + 1)
                    new_y = center_y + (next_y - center_y) * i / (skip_frames + 1)
                    results[frame_nb * (skip_frames + 1) + i].append(
                        (object_nb, new_x, new_y, object_class)
                    )
    logger.info("---Writing video")

    writer = FFmpegWriter(
        filename=output_filename,
        outputdict={
            "-pix_fmt": "rgb24",
            "-r": "%.02f" % fps,
            "-vcodec": "libx264",
            "-b": "5000000",
        },
    )

    font = cv2.FONT_HERSHEY_TRIPLEX
    for frame_nb, frame in enumerate(reader):
        detections_for_frame = results[frame_nb]
        for detection in detections_for_frame:
            if labels2icons is None:
                # write name of class
                cv2.putText(
                    frame,
                    f"{detection[0]}/{detection[3]}",
                    (int(detection[1]), int(detection[2]) + 5),
                    font,
                    2,
                    (0, 0, 255),
                    3,
                    cv2.LINE_AA,
                )
            else:
                # icons
                overlay_transparent(
                    frame,
                    labels2icons[detection[3]],
                    int(detection[1]) + 5,
                    int(detection[2]),
                )
                cv2.putText(
                    frame,
                    f"{detection[0]}",
                    (int(detection[1] + 46 + 5), int(detection[2]) + 42),
                    font,
                    1.2,
                    (0, 0, 0),
                    2,
                    cv2.LINE_AA,
                )

            if gps_data is not None:
                latitude = gps_data[frame_nb // fps]["Latitude"]
                longitude = gps_data[frame_nb // fps]["Longitude"]
                cv2.putText(
                    frame,
                    f"GPS:{latitude},{longitude}",
                    (10, frame.shape[0] - 30),
                    font,
                    2,
                    (0, 255, 0),
                    3,
                    cv2.LINE_AA,
                )

        frame = downscale_local_mean(frame, (downscale, downscale, 1)).astype(np.uint8)
        writer.writeFrame(frame[:, :, ::-1])

    writer.close()
    reader.video.release()

    logger.info("---finished writing video")


def resize_external_detections(detections: ndarray, ratio: float) -> ndarray:

    """Resize external detections with specific ratio.

    Args:
        detections (ndarray): the n-dimensionnal array of detections
        ratio (float): the ratio to resize detections

    Returns:
        detections (ndarray): the resized external detections
    """

    for detection_nb in range(len(detections)):
        detection = detections[detection_nb]
        if len(detection):
            detection = np.array(detection)[:, :-1]
            detection[:, 0] = (detection[:, 0] + detection[:, 2]) / 2
            detection[:, 1] = (detection[:, 1] + detection[:, 3]) / 2
            detections[detection_nb] = detection[:, :2] / ratio
    return detections


def write_tracking_results_to_file(
    results: ndarray, coord_mapping, output_filename: str
) -> None:

    """Writes the output results of a tracking in the following format:

    - frame
    - id
    - x_tl, y_tl, w=0, h=0
    - 4x unused=-1

    Args:
        results (ndarray): the output results
        coord_mapping (function): the mapping function for coordonates with ratio scaling
        output_filename (str): the name of the output file

    """

    with open(output_filename, "w") as output_file:
        for result in results:
            x, y = coord_mapping(result[2], result[3])
            output_file.write(
                "{},{},{},{},{},{},{},{},{},{}\n".format(
                    result[0] + 1,
                    result[1] + 1,
                    x,
                    y,
                    0,  # width
                    0,  # height
                    round(result[4], 2),
                    result[5],
                    -1,
                    -1,
                )
            )


def read_tracking_results(input_file: str) -> List[List]:

    """Read the input filename and interpret it as tracklets in this format : ``list[list,...]``.

    Args:
        input_file (str): the input file to read tracking results

    Returns:
        tracklets (List[List]): the tracking results stored in tracklets
    """

    raw_results = np.loadtxt(input_file, delimiter=",")
    if raw_results.ndim == 1:
        raw_results = np.expand_dims(raw_results, axis=0)
    tracklets = defaultdict(list)
    for result in raw_results:
        # Skip blank lines
        if result is None or len(result) == 0:
            continue
        frame_id = int(result[0])
        track_id = int(result[1])
        left, top, width, height = result[2:6]
        center_x = left + width / 2
        center_y = top + height / 2
        conf = result[6]
        class_id = int(result[7])
        tracklets[track_id].append((frame_id, center_x, center_y, conf, class_id))

    tracklets = list(tracklets.values())
    return tracklets


class FramesWithInfo:

    """Frames with information.

    Args:
        frames (ndarray): the list of frames
        output_shape (Optional[ndarray]): the shape of output frames
    """

    def __init__(self, frames: ndarray, output_shape: Optional[ndarray] = None):
        self.frames = frames
        if output_shape is None:
            self.output_shape = frames[0].shape[:-1][::-1]
        else:
            self.output_shape = output_shape
        self.end = len(frames)
        self.read_head = 0

    def __next__(self):
        if self.read_head < self.end:
            frame = self.frames[self.read_head]
            self.read_head += 1
            return frame
        else:
            raise StopIteration

    def __iter__(self):
        return self


class Display:

    """Display tracking.

    Args:
        on (bool): to activate display
        interactive (bool): to have interactive display. Set as default to ``True``.
    """

    def __init__(self, on: bool, interactive: bool = True):
        self.on = on
        self.fig, self.ax = plt.subplots()
        self.interactive = interactive
        if interactive:
            plt.ion()
        self.colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        self.legends = []
        self.plot_count = 0

    def display(self, trackers: Iterable) -> None:

        """To display trackers.

        Args:
            trackers (Iterable): the input trackers to display
        """

        something_to_show = False
        for tracker_nb, tracker in enumerate(trackers):
            if tracker.enabled:
                tracker.fill_display(self, tracker_nb)
                something_to_show = True

        self.ax.imshow(self.latest_frame_to_show)

        if len(self.latest_detections):
            self.ax.scatter(
                self.latest_detections[:, 0],
                self.latest_detections[:, 1],
                c="r",
                s=40,
            )

        if something_to_show:
            self.ax.xaxis.tick_top()
            plt.legend(handles=self.legends)
            self.fig.canvas.draw()
            if self.interactive:
                plt.show()
                while not plt.waitforbuttonpress():
                    continue
            else:
                plt.savefig(os.path.join("plots", str(self.plot_count)))
            self.ax.cla()
            self.legends = []
            self.plot_count += 1

    def update_detections_and_frame(
        self, latest_detections: ndarray[Any, dtype[generic]], frame: Mat
    ) -> None:

        """Update detections and frame.

        Args:
            latest_detections (ndarray): the lastest detections
            frame (Mat): the frame to update for resize
        """

        self.latest_detections = latest_detections
        self.latest_frame_to_show = cv2.cvtColor(
            cv2.resize(frame, self.display_shape), cv2.COLOR_BGR2RGB
        )



"""The ``make_dataset`` submodule provides only one main function to get annotations,
build annotations files for yolo and build dataset for training.

"""

from plasticorigins.training.data.data_processing import (
    get_annotations_from_files,
    get_annotations_from_db,
    find_img_ids_to_exclude,
)
from plasticorigins.training.data.data_processing import (
    generate_yolo_files,
    get_train_valid,
    build_yolo_annotations_for_images,
)
from pathlib import Path
import argparse
from argparse import Namespace


def main(args: Namespace) -> None:

    """Main Function to get annotations, build annotations files for yolo and build dataset for training.

    Args:
        args (argparse): list of arguments to build dataset for label mapping and training
    """

    data_dir = Path(args.data_dir)

    if args.bboxes_filename and args.images_filename:
        df_bboxes, df_images = get_annotations_from_files(
            data_dir, args.bboxes_filename, args.images_filename
        )
    elif args.password:
        print("getting annotations from db")
        df_bboxes, df_images = get_annotations_from_db(args.password)

    else:
        print("either a password must be set, or bbox and images filenames")
        return

    if args.exclude_img_folder:
        to_exclude = find_img_ids_to_exclude(args.exclude_img_folder)
    else:
        to_exclude = None

    # new
    yolo_filelist, cpos, cneg = build_yolo_annotations_for_images(
        data_dir,
        args.images_dir,
        df_bboxes,
        df_images,
        args.context_filters,
        args.quality_filters,
        args.limit_data,
        to_exclude,
    )

    # old
    # yolo_filelist, cpos, cneg = build_yolo_annotations_for_images_VM(data_dir, args.images_dir, args.bboxes_filename,
    #         df_bboxes, df_images, args.limit_data, to_exclude)

    print(
        f"found {cpos} valid annotations with images and {cneg} unmatched annotations"
    )

    train_files, val_files = get_train_valid(yolo_filelist, args.split)

    generate_yolo_files(data_dir, train_files, val_files)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Build dataset")
    parser.add_argument("--data-dir", type=str, help="path to main data folder")
    parser.add_argument("--images-dir", type=str, help="path to image folder")
    parser.add_argument("--password", type=str, help="password for connection to DB")
    parser.add_argument("--bboxes-filename", type=str, default="")
    parser.add_argument("--images-filename", type=str, default="")
    parser.add_argument(
        "--context-filters",
        type=str,
        help="context filters to apply",
        default=None,
    )
    parser.add_argument(
        "--quality-filters",
        type=str,
        help="quality filters to apply",
        default=None,
    )
    parser.add_argument("--split", type=float, default=0.85)
    parser.add_argument("--limit-data", type=int, default=0)
    parser.add_argument(
        "--exclude-img-folder",
        type=str,
        help="the path to the folder which contains images and annotations, from which we can find the img ids to exclude",
    )
    args = parser.parse_args()

    main(args)



"""The ``files`` submodule provides several functions for creating folders, files or loading files.

This submodule contains the following functions:

- ``create_unique_folder(base_folder:str, filename:str)`` : Creates a unique folder based on the filename and timestamp.
- ``download_from_url(url:str, filename:str, folder:str, logger:Logger)`` : Download a file and place it in the corresponding folder if it does
        not already exists.
- ``load_trash_icons(folder_path:str)`` : Loads all icons using cv2 format and returns a dict class.

"""

import datetime
import os
import os.path as op
from logging import Logger
from pathlib import Path, WindowsPath
from typing import Dict
from urllib.request import urlretrieve
import cv2


def create_unique_folder(base_folder: str, filename: str) -> str:

    """Creates a unique folder based on the filename and timestamp.

    Args:
        base_folder (str): main folder to store the unique folder
        filename (str): name of the file

    Returns:
        output_dir (str): new folder directory
    """

    folder_name = op.splitext(op.basename(filename))[0] + "_out_"
    folder_name += datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
    output_dir = op.join(base_folder, folder_name)

    if not op.isdir(output_dir):
        os.mkdir(output_dir)

    return output_dir


def download_from_url(
    url: str, filename: str, folder: str, logger: Logger
) -> WindowsPath:

    """Download a file and place it in the corresponding folder if it does
        not already exists. Useful for spaces demo, do not remove.

    Args:
        url (str): url for downloading the file
        filename (str): file to download
        folder (str): name of the folder
        logger (Logger): logger

    Returns:
        out_filename (WindowsPath): the path of the downloaded file
    """

    out_filename = Path(folder) / filename

    if not op.exists(out_filename):
        logger.info("---Downloading file...")
        urlretrieve(url, out_filename.resolve().as_posix())

    else:
        logger.info("---File already downloaded.")

    return out_filename


def load_trash_icons(folder_path: str) -> Dict:

    """Loads all icons using cv2 format and returns a dict class -> opened icon. Useful for spaces demo.

    Args:
        folder_path (str): path of the folder

    Returns:
        out_dict (Dict): Dictionnary of trash icons (resized)
    """

    folder_path = Path(folder_path)
    id_path = {
        "Fragment": folder_path
        / "fragment.png",  # 'Fragment',    #'Sheet / tarp / plastic bag / fragment',
        "Insulating": folder_path
        / "mousse.png",  # 'Insulating',  #'Insulating material',
        "Bottle": folder_path / "bouteille.png",  # 'Bottle',      #'Bottle-shaped',
        "Can": folder_path / "briquet.png",  # 'Can',         #'Can-shaped',
        "Drum": folder_path / "contenant.png",  # 'Drum',
        "Packaging": folder_path
        / "emballage.png",  # 'Packaging',   #'Other packaging',
        "Tire": folder_path / "pneu.png",  # 'Tire',
        "Fishing net": folder_path
        / "hamecon.png",  # 'Fishing net', #'Fishing net / cord',
        "Easily namable": folder_path / "chaussure.png",  # 'Easily namable',
        "Unclear": folder_path / "dechet.png",  # 'Unclear'
    }
    out_dict = {}

    for idx, path in id_path.items():
        img = cv2.imread(path.resolve().as_posix(), cv2.IMREAD_UNCHANGED)
        resized_img = cv2.resize(img, (100, 60), interpolation=cv2.INTER_AREA)
        out_dict[idx] = resized_img

    return out_dict



"""The ``losses`` submodule provides several loss functions to evaluate models.

This submodule contains the following class:

- ``FocalLoss`` : This class allows to compute the focal loss of a object detection model with several loss functions (train, test, focal).

This submodule contains the following functions:

- ``l1_loss(pred:torch.Tensor, gt:torch.Tensor, pos_inds:torch.Tensor)`` : Evaluate and compute the L1 loss on pred and gt based on pos_inds.
- ``sigmoid(x:torch.Tensor)`` : Compute the input Tensor with the sigmoid activation function.

"""

import torch
import torch.nn as nn
from torch.nn.functional import l1_loss as torch_l1_loss


class FocalLoss(nn.Module):

    """This class allows to compute the focal loss of a object detection model.

    Args:
        alpha (int): alpha coefficient for focal loss. Set as default to ``2``.
        beta (int): beta coefficient for focal loss. Set as default to ``4``.
        train (bool): ``True`` if we compute the focal loss for train phasis. In this case we use the ``_focal_loss`` method to compute the loss. Set as default to ``True``.
        centernet_output (bool): ``True`` if the model used for predict outputs is Centernet. Set as default to ``True``.
    """

    def __init__(
        self,
        alpha: int = 2,
        beta: int = 4,
        train: bool = True,
        centernet_output: bool = True,
    ):
        super().__init__()
        self.alpha = int(alpha)
        self.beta = int(beta)
        self.focal_loss = self._focal_loss if train else self._focal_loss_class_wise
        self.loss = self._train_loss if train else self._test_loss
        self.centernet_output = centernet_output

    def forward(self, pred: torch.Tensor, gt: torch.Tensor) -> float:

        """Compute the loss between predictions ``pred`` and the true values ``gt`` (ground truths).

        Args:
            pred (torch.Tensor): model predictions
            gt (torch.Tensor): ground truths

        Returns:
            The loss between predictions and ground truths.
        """

        return self.loss(pred, gt)

    def train_loss(self, pred: torch.Tensor, gt: torch.Tensor) -> float:

        """Evaluate and compute the training loss.

        Args:
            pred (torch.Tensor): model predictions
            gt (torch.Tensor): ground truths

        Returns:
            The training loss between predictions and ground truths.
        """

        if self.centernet_output:
            pred = pred[0]

        pred_centers = pred["hm"]
        gt_centers = gt[:, :-2, :, :]

        # pred_wh = pred['wh']
        # gt_wh = gt[:,-2:,:,:]

        pos_inds = gt_centers.eq(1).float()
        neg_inds = gt_centers.lt(1).float()

        return self.focal_loss(
            pred_centers, gt_centers, pos_inds, neg_inds
        )  # + 0.1 * _l1_loss(pred_wh, gt_wh, pos_inds)

    def test_loss(self, pred: torch.Tensor, gt: torch.Tensor) -> float:

        """Evaluate and compute the test loss.

        Args:
            pred (torch.Tensor): model predictions
            gt (torch.Tensor): ground truths

        Returns:
            The test loss between predictions and ground truths.
        """

        if self.centernet_output:
            pred = pred[0]

        pred_centers = pred["hm"]
        gt_centers = gt[:, :-2, :, :]

        pos_inds = gt_centers.eq(1).float()
        neg_inds = gt_centers.lt(1).float()

        return self._focal_loss_class_wise(pred_centers, gt_centers, pos_inds, neg_inds)

    def focal_loss(
        self,
        pred_hm: torch.Tensor,
        gt_hm: torch.Tensor,
        pos_inds: torch.Tensor,
        neg_inds: torch.Tensor,
    ) -> float:

        """Evaluate and compute the test loss. Modified focal loss. Exactly the same as CornerNet.
            Runs faster and costs a little bit more memory

        Args:
            pred_hm (torch.Tensor): predictions with size (batch x c x h x w)
            gt_hm (torch.Tensor): ground truths with size (batch x c x h x w)
            pos_inds (torch.Tensor): positive indices
            neg_inds (torch.Tensor): negative indices

        Returns:
            The focal loss between predictions and gt.
        """

        pred_hm = sigmoid(pred_hm)

        neg_weights = torch.pow(1 - gt_hm, self.beta)

        loss = 0.0

        pos_loss = torch.log(pred_hm) * torch.pow(1 - pred_hm, self.alpha) * pos_inds
        neg_loss = (
            torch.log(1 - pred_hm)
            * torch.pow(pred_hm, self.alpha)
            * neg_weights
            * neg_inds
        )

        num_pos = pos_inds.float().sum()
        pos_loss = pos_loss.sum()
        neg_loss = neg_loss.sum()

        if num_pos == 0:
            loss = loss - neg_loss
        else:
            loss = loss - (pos_loss + neg_loss) / num_pos

        return loss

    def focal_loss_class_wise(
        self,
        pred_hm: torch.Tensor,
        gt_hm: torch.Tensor,
        pos_inds: torch.Tensor,
        neg_inds: torch.Tensor,
    ) -> torch.Tensor:

        """Evaluate and compute the test loss. Modified focal loss. Exactly the same as CornerNet.
            Runs faster and costs a little bit more memory.

        Args:
            pred_hm (torch.Tensor): predictions with size (batch x c x h x w)
            gt_hm (torch.Tensor): ground truths with size (batch x c x h x w)
            pos_inds (torch.Tensor): positive indices
            neg_inds (torch.Tensor): negative indices

        Returns:
            The focal loss Tensor between predictions and gt.
        """

        pred_hm = sigmoid(pred_hm)

        neg_weights = torch.pow(1 - gt_hm, self.beta)

        pos_loss = torch.log(pred_hm) * torch.pow(1 - pred_hm, self.alpha) * pos_inds
        neg_loss = (
            torch.log(1 - pred_hm)
            * torch.pow(pred_hm, self.alpha)
            * neg_weights
            * neg_inds
        )

        num_pos = torch.sum(pos_inds.float(), dim=(0, 2, 3))
        pos_loss = torch.sum(pos_loss, dim=(0, 2, 3))
        neg_loss = torch.sum(neg_loss, dim=(0, 2, 3))

        loss = torch.zeros(size=(len(num_pos),))

        for i, num_poss_class in enumerate(num_pos):

            if num_poss_class == 0:
                loss[i] = loss[i] - neg_loss[i]
            else:
                loss[i] = loss[i] - (pos_loss[i] + neg_loss[i]) / num_poss_class

        return loss


def sigmoid(x: torch.Tensor) -> torch.Tensor:

    """Compute the input Tensor with the sigmoid activation function.

    Args:
        x (torch.Tensor): the input Tensor

    Returns:
        y (torch.Tensor): the output Tensor which corresponds to the application of the sigmoid function on the input
    """

    y = torch.clamp(x.sigmoid_(), min=1e-4, max=1 - 1e-4)

    return y


def l1_loss(
    pred: torch.Tensor, gt: torch.Tensor, pos_inds: torch.Tensor
) -> torch.Tensor:

    """Evaluate and compute the L1 loss on predictions `pred` and ground truths `gt` based on ``pos_inds``. L1 Loss Function is used to minimize the error which is
        the sum of the all the absolute differences between the true value and the predicted value.

    Args:
        pred (torch.Tensor): predictions with size (batch x c x h x w)
        gt (torch.Tensor): ground truths with size (batch x c x h x w)
        pos_inds (torch.Tensor): positive indices

    Returns:
        The L1 Loss Tensor between predictions and ground truths.
    """

    pos_inds = torch.sum(pos_inds, axis=1).ge(1).unsqueeze(1).expand_as(pred)

    return torch_l1_loss(pred[pos_inds], gt[pos_inds], size_average=False) / (
        pos_inds.float().sum() + 1e-4
    )



id_categories = {
    0: "Fragment",  # 'Sheet / tarp / plastic bag / fragment',
    1: "Insulating",  # 'Insulating material',
    2: "Bottle",  # 'Bottle-shaped',
    3: "Can",  # 'Can-shaped',
    4: "Drum",
    5: "Packaging",  # 'Other packaging',
    6: "Tire",
    7: "Fishing net",  # 'Fishing net / cord',
    8: "Easily namable",
    9: "Unclear",
}


class DotDict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


config_track = DotDict(
    {
        "upload_folder": "/tmp",
        "confidence_threshold": 0.004,
        "detection_threshold": 0.3,
        "downsampling_factor": 4,
        "noise_covariances_path": "data/tracking_parameters",
        "output_shape": (960, 544),
        "skip_frames": 3,  # 3
        "arch": "mobilenet_v3_small",
        "device": "cpu",
        "detection_batch_size": 1,
        "display": 0,
        "kappa": 5,  # 7
        "tau": 3,  # 4
        "crop": False,
    }
)

config_track_yolo = DotDict(
    {
        "yolo_conf_thrld": 0.35,
        "yolo_iou_thrld": 0.5,
        "upload_folder": "/tmp",
        "confidence_threshold": 0.004,
        "downsampling_factor": 4,
        "noise_covariances_path": "data/tracking_parameters",
        "url_model_yolo": "https://github.com/surfriderfoundationeurope/IA_Pau/releases/download/v0.1/yolov5.pt",
        "file_model_yolo": "yolov5.pt",
        "output_shape": (768, 768),
        "skip_frames": 3,  # 3
        "arch": "yolo",
        "device": "cpu",
        "detection_batch_size": 1,
        "display": 0,
        "kappa": 5,  # 7
        "tau": 3,  # 4
        "crop": True,
    }
)


logging_config = {
    "version": 1,
    "formatters": {
        "default": {"format": "[%(asctime)s] %(levelname)s in %(module)s: %(message)s"}
    },
    "handlers": {
        "wsgi": {
            "class": "logging.StreamHandler",
            "stream": "ext://flask.logging.wsgi_errors_stream",
            "formatter": "default",
        },
        "file": {
            "class": "logging.FileHandler",
            "formatter": "default",
            "filename": "errors.log",
        },
    },
    "root": {"level": "INFO", "handlers": ["wsgi", "file"]},
}



"""The ``postprocess_and_count_tracks`` submodule provides several functions to post process and count tracks.

The module allows the user to :

- filter tracks
- process object class and confidences
- count objects on the frames

This submodule contains the following functions :

- ``compute_moving_average(tracklet:List[List], kappa:int)`` : Computing the moving average of the tracks depending on parameter ``kappa``.
- ``count_objects(input_json:Dict, class_dict:Dict)`` : Counting trashs from the ``input_json`` object.
- ``filter_by_nb_consecutive_obs(tracklets:List[List], kappa:int, tau:int)`` : Filters the tracks depending on parameters ``kappa`` and ``tau`` by consecutive observations.
- ``filter_tracks(tracklets:List[List], kappa:int, tau:int)`` : Filters the tracks depending on parameters ``kappa`` and ``tau``.
- ``postprocess_for_api(results:List[Tuple], class_dict:Dict=defaultdict(lambda: "fragment"))`` : Converts tracking results into json object for API.
- ``process_class_and_confidences(class_confs:List[Tuple[int,float]])`` : Finds the majority and most confident class from list `class_confs`.
- ``threshold(tracklets:List[List], tau:int)`` : Filters the tracks depending on parameter ``tau``.

"""

import argparse
import json
from collections import defaultdict
from typing import List, Tuple, Dict
import numpy as np
from numpy import array
from scipy.signal import convolve

from plasticorigins.tracking.utils import (
    read_tracking_results,
    write_tracking_results_to_file,
)


def filter_by_nb_consecutive_obs(tracklets: List[List], kappa: int, tau: int) -> List:

    """Filters the tracks depending on parameters ``kappa`` and ``tau`` by consecutive observations.

    Args:
        tracklets (List[List]): list of tracks to filter
        kappa (int): size of the moving average window
        tau (int): minimum length of tracklet

    Returns:
        results (List[Tuple]) : raw filtered tracks of minimum length ``tau``
    """

    new_tracklets = []

    for tracklet in tracklets:
        new_tracklet = []
        density_fill = compute_moving_average(tracklet, kappa=kappa)
        for (observation, density_fill_value) in zip(tracklet, density_fill):
            if density_fill_value > 0.6:
                new_tracklet.append(observation)
        new_tracklets.append(new_tracklet)

    return threshold(new_tracklets, tau)


def filter_tracks(tracklets: List[List], kappa: int, tau: int) -> List[Tuple]:

    """Filters the tracks depending on parameters ``kappa`` and ``tau``.

    Args:
        tracklets (List[List]): list of tracks to filter
        kappa (int): size of the moving average window
        tau (int): minimum length of tracklet

    Returns:
        results (List[Tuple]) : raw filtered tracks
    """

    if not kappa == 1:
        tracks = filter_by_nb_consecutive_obs(tracklets, kappa, tau)
    else:
        tracks = tracklets

    results = []

    for tracker_nb, dets in enumerate(tracks):
        for det in dets:
            results.append((det[0], tracker_nb, det[1], det[2], det[3], det[4]))

    results = sorted(results, key=lambda x: x[0])

    return results


def process_class_and_confidences(
    class_confs: List[Tuple[int, float]]
) -> Tuple[int, float]:

    """Finds the majority and most confident class from list `class_confs`.

    Args:
        class_confs (List[Tuple[int,float]]): the list of class ids and confidences such as ``[(class_id, confidence), ...]``

    Returns:
        best class (Tuple[int,float]) : The best class and its associated confidence
    """

    d = defaultdict(lambda: (0, 0.0))

    for (cls, conf) in class_confs:
        d[cls] = (d[cls][0] + 1, d[cls][1] + conf)

    best_class = sorted(d.items(), key=lambda v: v[1][0] + v[1][1])[-1]

    return best_class[0], round(best_class[1][1] / best_class[1][0], 2)


def postprocess_for_api(
    results: List[Tuple], class_dict: Dict = defaultdict(lambda: "fragment")
) -> Dict:

    """Converts tracking results into json object for API.

    Args:
        results (List[Tuple]) : raw filtered tracks
        class_dict (Dict): the dictionnary of object classes

    Returns:
        The detected trashs dictionnary (json object format) ``{"detected_trash" : result_list}`` with labels and average confidences in `result_list`
    """

    result_list = []
    id_list = {}

    for res in results:
        frame_number = res[0]
        box = [
            round(res[2], 1),
            round(res[3], 1),
            round(res[2], 1),
            round(res[3], 1),
        ]
        id = res[1]
        conf = round(res[4], 2)
        classname = class_dict[res[5]]
        # if the id is not already is the results, add a new jsonline
        if id not in id_list:
            id_list[id] = len(result_list)
            result_list.append(
                {
                    "label": classname,
                    "id": id,
                    "frame_to_box": {str(frame_number): box},
                    "frame_to_class_conf": {str(frame_number): (res[5], conf)},
                }
            )
        # otherwise, retrieve the jsonline and append the box
        else:
            result_list[id_list[id]]["frame_to_box"][str(frame_number)] = box
            result_list[id_list[id]]["frame_to_class_conf"][str(frame_number)] = (
                res[5],
                conf,
            )

    # Finally, collapse the confidence and class
    for res in result_list:
        classid, avg_conf = process_class_and_confidences(
            res.pop("frame_to_class_conf").values()
        )
        res["avg_conf"] = avg_conf
        # update the label
        res["label"] = class_dict[classid]

    return {"detected_trash": result_list}


def count_objects(input_json: Dict, class_dict: Dict) -> Dict:

    """Counting trashs from the ``input_json`` object.

    Args:
        input_json (Dict) : the detected trashs dictionnary ``{"detected_trash" : result_list}``
        class_dict (Dict): the dictionnary of object classes

    Returns:
        The dictionnary of trash proportions.
    """

    results = {v: 0 for v in class_dict.values()}
    total = 0

    for trash in input_json["detected_trash"]:
        results[trash["label"]] += 1
        total += 1

    if total == 0:
        total = 1

    return {k + f": {str(v)}": v / total for k, v in results.items()}


def write(results: List[Tuple], output_name: str):

    """Writes the results in two files:
    - tracking in a ``xxx_track.txt`` format  ``(frame, id, box_x, box_y, ...)``
    - the number of detected objects in a separate file ``xxx_count.txt``

    Args:
        results (List[Tuple]) : raw filtered tracks
        output_name (str): the name of the output file for tracking and counting
    """

    output_tracks_filename = output_name.split(".")[0] + "_tracks.txt"
    write_tracking_results_to_file(
        results, lambda x, y: (x, y), output_filename=output_tracks_filename
    )

    with open(output_name.split(".")[0] + "_count.txt", "w") as out_file:
        if len(results):
            out_file.write(f"{max(result[1]+1 for result in results)}")
        else:
            out_file.write("0")


def threshold(tracklets: List[List], tau: int) -> List:

    """Filters the tracks depending on parameter ``tau``.

    Args:
        tracklets (List[List]): list of tracks to filter
        tau (int): minimum length of tracklet

    Returns:
        results (List) : raw filtered tracks
    """

    return [tracklet for tracklet in tracklets if len(tracklet) > tau]


def compute_moving_average(tracklet: List[List], kappa: int) -> array:

    """Computing the moving average of the tracks depending on parameter ``kappa``.

    Args:
        tracklet (List[List]): list of tracks to filter
        kappa (int): size of the moving average window

    Returns:
        density_fill (array) : moving average resulting from convolution between observation points and ``kappa``
    """

    if len(tracklet) == 0 or len(tracklet[0]) == 0:
        return tracklet

    pad = (kappa - 1) // 2
    observation_points = np.zeros(tracklet[-1][0] - tracklet[0][0] + 1)
    first_frame_id = tracklet[0][0] - 1

    for observation in tracklet:
        frame_id = observation[0] - 1
        observation_points[frame_id - first_frame_id] = 1

    density_fill = convolve(observation_points, np.ones(kappa) / kappa, mode="same")

    if pad > 0 and len(observation_points) >= kappa:
        density_fill[:pad] = density_fill[pad : 2 * pad]
        density_fill[-pad:] = density_fill[-2 * pad : -pad]

    density_fill = observation_points * density_fill

    return density_fill[density_fill > 0]


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_name", type=str)
    parser.add_argument("--kappa", type=int)
    parser.add_argument("--tau", type=int)
    parser.add_argument("--output_type", type=str, default="api")
    args = parser.parse_args()

    tracklets = read_tracking_results(args.input_file)
    filtered_results = filter_tracks(tracklets, args.kappa, args.tau)
    if args.output_type == "api":
        output = postprocess_for_api(filtered_results)
        with open(args.output_name, "w") as f:
            json.dump(output, f)
    else:
        write(filtered_results, args.output_name)
